/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/sysctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <netdb.h>
#include <resolv.h>
#include <string.h>
#include <stdlib.h>

#define SUCCESS 0
#define ANY 0
#define YES 1
#define NO  0

static struct addrinfo pai;
static u_short port;
static char *proto;

static int translate = NO;
static struct in6_addr faith_prefix = IN6ADDR_ANY_INIT;

static char in_addrany[] = { 0, 0, 0, 0};
static char in6_addrany[] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
static char in_loopback[] = { 127, 0, 0, 1}; 
static char in6_loopback[] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 };

struct sockinet {
	u_char	si_len;
	u_char	si_family;
	u_short	si_port;
};

/*
 * Note that a_off will be dynamically adjusted so that to be consistent
 * with the definition of sockaddr_in{,6}.
 * The value presented below is just a guess.
 */
static struct afd {
	int a_af;
	int a_addrlen;
	int a_socklen;
	int a_off;
	char *a_addrany;
	char *a_loopback;	
} afdl [] = {
	/* first entry is linked last... */
	{PF_INET, sizeof(struct in_addr),
	 sizeof(struct sockaddr_in), 4 /*XXX*/,
	 in_addrany, in_loopback},
	{PF_INET6, sizeof(struct in6_addr),
	 sizeof(struct sockaddr_in6), 8 /*XXX*/,
	 in6_addrany, in6_loopback},
	{0, 0, 0, 0, NULL, NULL},
};

#define N_INET  0
#define N_INET6 1

static int get_name __P((const char *, struct afd *,
			  struct addrinfo **, char *));
static int get_addr __P((const char *, struct afd *, struct addrinfo **));
static int get_addr_4or6 __P((const char *, struct addrinfo **));
static int isnumber __P((const char *));
	
static char *ai_errlist[] = {
	"success.",
	"address family for hostname not supported.",	/* EAI_ADDRFAMILY */
	"temporary failure in name resolution.",	/* EAI_AGAIN      */
	"invalid value for ai_flags.",		       	/* EAI_BADFLAGS   */
	"non-recoverable failure in name resolution", 	/* EAI_FAIL       */
	"ai_family not supported.",			/* EAI_FAMILY     */
	"memory allocation failure.", 			/* EAI_MEMORY     */
	"no address associated with hostname.", 	/* EAI_NODATA     */
	"hostname nor servname provided, or not known.",/* EAI_NONAME     */
	"servname not supported for ai_socktype.",	/* EAI_SERVICE    */
	"ai_socktype not supported.", 			/* EAI_SOCKTYPE   */
	"system error returned in errno.", 		/* EAI_SYSTEM     */
	"invalid value for hints.",			/* EAI_BADHINTS	  */
	"resolved protocol is unknown.",		/* EAI_PROTOCOL   */
	"unknown error.", 				/* EAI_MAX        */
};

#define GET_CANONNAME(ai, str) \
if (pai.ai_flags & AI_CANONNAME) {\
	if ((ai)->ai_canonname = (char *)malloc(strlen(str) + 1)) {\
		strcpy((ai)->ai_canonname, (str));\
	} else {\
		error = EAI_MEMORY;\
		goto free;\
	}\
}

#define GET_AI(ai, afd, addr) {\
	char *p;\
	if (((ai) = (struct addrinfo *)malloc(sizeof(struct addrinfo) +\
					      ((afd)->a_socklen)))\
	    == NULL) goto free;\
	*(ai) = pai;\
	(ai)->ai_addr = (struct sockaddr *)((ai) + 1);\
	memset((ai)->ai_addr, 0, (afd)->a_socklen);\
	(ai)->ai_addr->sa_len = (ai)->ai_addrlen = (afd)->a_socklen;\
	(ai)->ai_addr->sa_family = (ai)->ai_family = (afd)->a_af;\
	((struct sockinet *)(ai)->ai_addr)->si_port = port;\
	p = (char *)((ai)->ai_addr);\
	memcpy(p + (afd)->a_off, (addr), (afd)->a_addrlen);\
}

#define ERR(err) { error = (err); goto bad; }

char *
gai_strerror(ecode)
	int ecode;
{
	if (ecode < 0 || ecode > EAI_MAX)
		ecode = EAI_MAX;
	return ai_errlist[ecode];
}

void
freeaddrinfo(ai)
	struct addrinfo *ai;
{
	struct addrinfo *next;

	do {
		next = ai->ai_next;
		if (ai->ai_canonname) free(ai->ai_canonname);
		/* no need to free(ai->ai_addr) */
		free(ai);
	} while (ai = next);
}

static int
isnumber(p)
	const char *p;
{
	char *q = (char *)p;
	while(*q) {
		if (! isdigit(*q)) return NO;
		q++;
	}
	return YES;
}

int
getaddrinfo(hostname, servname, hints, res)
	const char *hostname, *servname;
	const struct addrinfo *hints;
	struct addrinfo **res;
{
	struct addrinfo *top = NULL;
	struct addrinfo *cur = NULL;
	int i, error = 0;
	char pton[sizeof(struct in6_addr)];
	static int firsttime = 1;

	/* dynamically adjust a_off */
	if (firsttime) {
		struct afd *p;
		u_char *q;
		struct sockaddr_in sin;
		struct sockaddr_in6 sin6;

		for (p = &afdl[0]; p->a_af; p++) {
			switch (p->a_af) {
			case PF_INET:
				q = (u_char *)&sin.sin_addr.s_addr;
				p->a_off = q - (u_char *)&sin;
				break;
			case PF_INET6:
				q = (u_char *)&sin6.sin6_addr.s6_addr8;
				p->a_off = q - (u_char *)&sin6;
				break;
			default:
				break;
			}
		}

		/* translator hack */
		q = getenv("GAI");
		if (q && inet_pton(AF_INET6, q, &faith_prefix) == 1)
			translate = YES;
	
		firsttime = 0;
	}

	/* initialize file static vars */
	pai.ai_flags = 0;
	pai.ai_family = ANY;
	pai.ai_socktype = PF_UNSPEC;
	pai.ai_protocol = ANY;
	pai.ai_addrlen = 0;
	pai.ai_canonname = NULL;
	pai.ai_addr = NULL;
	pai.ai_next = NULL;
	port = ANY;
	proto = NULL;
	
	if (hostname == NULL && servname == NULL)
		return EAI_NONAME;
	if (hints) {
		/* error check for hints */
		if (hints->ai_addrlen || hints->ai_canonname ||
		    hints->ai_addr || hints->ai_next)
			ERR(EAI_BADHINTS); /* xxx */
		if (hints->ai_flags & ~AI_MASK) ERR(EAI_BADFLAGS);
		if (hints->ai_family != PF_UNSPEC &&
		    hints->ai_family != PF_INET   &&
		    hints->ai_family != PF_INET6)
			ERR(EAI_FAMILY);
		pai = *hints;
		switch (pai.ai_socktype) {
		case ANY:
			switch (pai.ai_protocol) {
			case ANY:
				break;
			case IPPROTO_UDP:
				pai.ai_socktype = SOCK_DGRAM;
				break;
			case IPPROTO_TCP:
				pai.ai_socktype = SOCK_STREAM;
				break;
			default:
				pai.ai_socktype = SOCK_RAW;
				break;
			}
			break;
		case SOCK_RAW:
			break;
		case SOCK_DGRAM:
			if (pai.ai_protocol != IPPROTO_UDP &&
			    pai.ai_protocol != ANY)
				ERR(EAI_BADHINTS);
			pai.ai_protocol = IPPROTO_UDP;
			break;
		case SOCK_STREAM:
			if (pai.ai_protocol != IPPROTO_TCP &&
			    pai.ai_protocol != ANY)
				ERR(EAI_BADHINTS);
			pai.ai_protocol = IPPROTO_TCP;
			break;
		default:
			ERR(EAI_SOCKTYPE);
			break;
		}
	}

	/*
	 * service port
	 */

	if (servname) {
		if (isnumber(servname)) {
			if (pai.ai_socktype == ANY) ERR(EAI_SERVICE);
			port = htons(atoi(servname));
		} else {
			struct servent *sp;
			switch (pai.ai_socktype) {
			case ANY:
				proto = NULL;
				break;
			case SOCK_DGRAM:
				proto = "udp";
				break;
			case SOCK_STREAM:
				proto = "tcp";
				break;
			default:
				fprintf(stderr, "panic!\n");
				break;
			}
			if ((sp = getservbyname(servname, proto)) == NULL)
				ERR(EAI_SERVICE);
			port = sp->s_port;
			if (pai.ai_socktype == ANY)
				if (strcmp(sp->s_proto, "udp") == 0) {
					pai.ai_socktype = SOCK_DGRAM;
					pai.ai_protocol = IPPROTO_UDP;
				} else if (strcmp(sp->s_proto, "tcp") == 0) {
					pai.ai_socktype = SOCK_STREAM;
					pai.ai_protocol = IPPROTO_TCP;
				} else
					ERR(EAI_PROTOCOL);
		}
	}
	
	/*
	 * passive socket
	 */
	if (pai.ai_flags & AI_PASSIVE) {
		struct afd *afd;

		for (afd = &afdl[0]; afd->a_af; afd++) {
			if (!(pai.ai_family == PF_UNSPEC
			   || pai.ai_family == afd->a_af)) {
				continue;
			}

			if (!top) {
				GET_AI(top, afd, afd->a_addrany);
				cur = top;
			} else {
				GET_AI(cur->ai_next, afd, afd->a_addrany);
				cur = cur->ai_next;
			}
			/* xxx meaningless? 
			 * GET_CANONNAME(top, "anyaddr");
			 */
		}
		goto good;
	}

	/*
	 * non-passive socket
	 */

	/* hostname as anyaddr */
	if (hostname == NULL) {
		struct afd *afd;

		for (afd = &afdl[0]; afd->a_af; afd++) {
			if (!(pai.ai_family == PF_UNSPEC
			   || pai.ai_family == afd->a_af)) {
				continue;
			}

			if (!top) {
				GET_AI(top, afd, afd->a_loopback);
				GET_CANONNAME(top, "localhost");
				cur = top;
			} else {
				GET_AI(cur->ai_next, afd, afd->a_loopback);
				GET_CANONNAME(cur->ai_next, "localhost");
				cur = cur->ai_next;
			}
		}
		goto good;
	}
	
	/* hostname as numeric name */
	for (i = 0; afdl[i].a_af; i++) {
		if (inet_pton(afdl[i].a_af, hostname, pton)) {
			u_long v4a;
			u_char pfx;

			switch (afdl[i].a_af) {
			case AF_INET:
				v4a = ((struct in_addr *)pton)->s_addr;
				if (IN_MULTICAST(v4a) || IN_EXPERIMENTAL(v4a))
					pai.ai_flags &= ~AI_CANONNAME;
				v4a >>= IN_CLASSA_NSHIFT;
				if (v4a == 0 || v4a == IN_LOOPBACKNET)
					pai.ai_flags &= ~AI_CANONNAME;
				break;
			case AF_INET6:
				pfx = ((struct in6_addr *)pton)->s6_addr8[0];
				if (pfx == 0 || pfx == 0xfe || pfx == 0xff)
					pai.ai_flags &= ~AI_CANONNAME;
				break;
			}
			
			if (pai.ai_family == afdl[i].a_af ||
			    pai.ai_family == PF_UNSPEC) {
				if (! (pai.ai_flags & AI_CANONNAME)) {
					GET_AI(top, &afdl[i], pton);
					goto good;
				}
				/*
				 * if AI_CANONNAME and if reverse lookup
				 * fail, return ai anyway to pacify
				 * calling application.
				 */
				get_name(pton, &afdl[i], &top, pton);
				goto good;
			} else 
				ERR(EAI_BADHINTS);
		}
	}

	/* hostname as alphabetical name */
	switch (pai.ai_family) {
	case PF_UNSPEC:
		error = get_addr_4or6(hostname, &top);
		break;
	case PF_INET:
		error = get_addr(hostname, &afdl[N_INET], &top);
		break;
	case PF_INET6:
		error = get_addr(hostname, &afdl[N_INET6], &top);
		break;
	}
	if (error == 0) {
		if (top) {
		good:
			*res = top;
			return SUCCESS;
		} else 
			error = EAI_FAIL;
	}
 free:
	if (top) freeaddrinfo(top);
 bad:
	*res = NULL;
	return error;
}

static int
get_name(addr, afd, res, numaddr)
	const char *addr;
	struct afd *afd;
	struct addrinfo **res;
	char *numaddr;
{
	struct hostent *hp;
	struct addrinfo *cur;
	char *p, *nm;
	int msize, len, error = 0;
	
	if ((hp = gethostbyaddr(addr, afd->a_addrlen, afd->a_af)) &&
	    hp->h_name && hp->h_name[0] && hp->h_addr_list[0]) {
		GET_AI(cur, afd, hp->h_addr_list[0]);
		GET_CANONNAME(cur, hp->h_name);
	} else
		GET_AI(cur, afd, numaddr);
	
	*res = cur;
	return SUCCESS;
 free:
	if (cur) freeaddrinfo(cur);
 bad:
	*res = NULL;
	return error;
}

static int
get_addr(hostname, afd, res)
	const char *hostname;
	struct afd *afd;
	struct addrinfo **res;
{
	struct hostent *hp;
	struct addrinfo *top = NULL, *cur, *prev;
	struct afd *transafd = afd;
	char *p, *nm;
	int i, len, error = 0;

	if (translate && afd->a_af == AF_INET)
		transafd = &afdl[N_INET6];

	if ((hp = gethostbyname2(hostname, afd->a_af)) == NULL)
		ERR(EAI_NODATA);

	if ((hp->h_name == NULL) || (hp->h_name[0] == 0) ||
	    (hp->h_addr_list[0] == NULL))
		ERR(EAI_FAIL);
	
	for (i = 0; hp->h_addr_list[i]; i++) {
		GET_AI(cur, transafd, hp->h_addr_list[i]);
		if (translate  && afd->a_af == AF_INET) {
			struct in6_addr *in6 = &((struct sockaddr_in6 *)cur->ai_addr)->sin6_addr;
			/* faith prefix len is implicitly defined to be 12 */
			memcpy(&(in6->s6_addr32[0]), &faith_prefix, 12);
			memcpy(&(in6->s6_addr32[3]), hp->h_addr_list[i],
			      sizeof(struct in_addr));
		}
		if (top)
			prev->ai_next = cur;
		else {
			top = cur;
			GET_CANONNAME(top, hp->h_name);
		}
		prev = cur;
	}

	*res = top;
	return SUCCESS;
 free:
	if (top) freeaddrinfo(top);
 bad:
	*res = NULL;
	return error;
}

static int
get_addr_4or6(hostname, res)
	const char *hostname;
	struct addrinfo **res;
{
	
	int i, error = 0;
	struct addrinfo *cur = NULL, *top = NULL, *tmp;
	
	for (i = 0; afdl[i].a_af; i++) {
		error = get_addr(hostname, &afdl[i], &cur);
		switch (error) {
		case SUCCESS:
			if (top) {
				tmp = cur;
				while (tmp->ai_next)
					tmp = tmp->ai_next;
				tmp->ai_next = top;
			} 
			top = cur;
			break;
		case EAI_NODATA:
			continue;
		case EAI_FAIL:
		case EAI_MEMORY:	
			goto free;
		}
	}

	*res = top;
	return SUCCESS;
 free:
	if (top) freeaddrinfo(top);
 bad:
	*res = NULL;
	return error;
}
