/*
 * This program is an implementation of the ISAKMP Internet Standard.
 * Copyright (C) 1997 Angelos D. Keromytis.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * 
 * This code was written while the author was in Greece, in May/June
 * 1997.
 *
 * You may contact the author by:
 *   e-mail: angelos@dsl.cis.upenn.edu
 *  US-mail: Angelos D. Keromytis
 *           Distributed Systems Lab
 *           Computer and Information Science Department
 *           University of Pennsylvania
 *           Moore Building
 *           200 South 33rd Street
 *           Philadelphia, PA 19104	   
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "constants.h"
#ifdef INET6
#include "sockaddr6.h"
#endif /* INET6 */

#if defined(HYDRANGEA) || defined(LIBRA)
#define NETKEY
#endif

#ifdef NETKEY
#include <netkey/key.h>
#ifdef HYDRANGEA
#include <netinet6/in6.h>
#endif  /* HYDRANGEA */
#endif /* NETKEY */

/*
 * This file implements all the kernel communicating routines.
 * XXX Eventually it should use PF_KEY or what kernel API is available
 * XXX but for now this will have to do.
 */

/*
 * Handle a kernel request. Supposedly, there's a message in
 * the kernelsock socket.
 */
void
#ifdef INET6
kernel_handle(int kernelfd, int sock4, int sock6)
#else /* INET6 */
kernel_handle(int kernelfd, int sock)
#endif /* INET6 */
{
    u_char buffer[1024];
    int n;
#ifdef INET6
    struct sockaddr6 sin;
#else /* INET6 */
    struct sockaddr_in sin;
#endif /* INET6 */
    int goal;
    
    n = read(kernelfd, buffer, 1024);
    if (n == -1)
    {
	log(1, "read() failed in kernel_handle()", 0, 0, 0);
	return;
    }

#ifdef DEBUG
    log(0, "read %d bytes from \"kernel\" socket", n, 0, 0);
#endif

#ifdef NETKEY
    {
	u_char *ptr;
	struct key_msghdr *msg;
	struct sockaddr6 *src, *dst;

	ptr = buffer;

	if (n < sizeof(*msg))
		return;
	msg = (struct key_msghdr *) ptr;
	n -= sizeof(*msg);
	ptr += sizeof(*msg);
	if (msg->key_msgtype != KEY_ACQUIRE)
		return;
	switch (msg->type) {
		case KEY_TYPE_AH:
			goal = GOAL_AUTHENTICATE;
			break;
		case KEY_TYPE_ESP:
			goal = GOAL_ENCRYPT;
			break;
		default:
			return;
	}
	if (n < sizeof(src->sa6_len))
		return;
	src = (struct sockaddr6 *) ptr;
	if (n < src->sa6_len)
		return;
	if (n < sizeof(src->sa6_len))
		return;
	n -= src->sa6_len;
	ptr += src->sa6_len;
	dst = (struct sockaddr6 *) ptr;
	if (n < dst->sa6_len)
		return;
	if (dst->sa6_family == AF_INET6)
		if (IN6_IS_ADDR_MULTICAST(&dst->sa6_sin6.sin6_addr))
			return;
	bcopy(dst, &sin, dst->sa6_len);
	sin.sa6_sin6.sin6_port = htons(PORT);
    }
#ifdef DEBUG
{
    char ntop_buf[INET6_ADDRSTRLEN];
    SA6_NTOP(sin, ntop_buf, sizeof(ntop_buf));
    log(0, "initiating exchange with [%s], port %d, goal %d", 
	ntop_buf, get_port(sin), goal);
}
#endif
#else /* NETKEY */
    bcopy(buffer, &sin, sizeof(sin));
    if (sin.sa6_family == AF_INET6)
	sin.sa6_sin6.sin6_flowinfo = 0;
#ifdef INET6
    bcopy(buffer + sin.sa6_len, &goal, sizeof(goal));

#ifdef DEBUG
{
    char ntop_buf[INET6_ADDRSTRLEN];
    SA6_NTOP(sin, ntop_buf, sizeof(ntop_buf));
    log(0, "initiating exchange with [%s], port %d, goal %d", 
	ntop_buf, get_port(sin), goal);
}
#endif
#else /* INET6 */
    bcopy(buffer + sizeof(sin), &goal, sizeof(goal));
    
#ifdef DEBUG
    log(0, "initiating exchange with [%s], port %d, goal %d", 
	inet_ntoa(sin.sin_addr), ntohs(sin.sin_port), goal);
#endif
#endif /* INET6 */
#endif /* NETKEY */

#ifdef INET6
    ipsecdoi_initiate((sin.sa6_family == AF_INET ? sock4 : sock6),
		      sin, goal, KEY_OAKLEY);
#else /* INET6 */
    ipsecdoi_initiate(sock, sin, goal, KEY_OAKLEY);
#endif /* INET6 */
}
