/*
 * This program is an implementation of the ISAKMP Internet Standard.
 * Copyright (C) 1997 Angelos D. Keromytis.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * 
 * This code was written while the author was in Greece, in May/June
 * 1997.
 *
 * You may contact the author by:
 *   e-mail: angelos@dsl.cis.upenn.edu
 *  US-mail: Angelos D. Keromytis
 *           Distributed Systems Lab
 *           Computer and Information Science Department
 *           University of Pennsylvania
 *           Moore Building
 *           200 South 33rd Street
 *           Philadelphia, PA 19104	   
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <netdb.h>
#include <unistd.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include "constants.h"
#include "state.h"
#include "defs.h"

#if defined(HYDRANGEA) || defined(LIBRA)
#define NETKEY
#endif

#ifdef NETKEY
#ifdef LIBRA
#include <netkey/key.h>
#include <ptcpip/ipsec.h>
#else /* HYDRANGEA */
#ifdef  INET6
#include <netinet6/in6.h>
#endif /* INET6 */
#include <netkey/key.h>
#include <netinet6/ah.h>
#include <netinet6/esp.h>
#endif
#endif /* NETKEY */

/*
 *  Server main loop and socket initialization routines.
 */

/*
 * Initialize the kernel socket.
 * XXX For now, just get a port.
 */
static int
write_register(fd, type)
	int fd;
	int type;
{
	struct key_msghdr msg;

	bzero(&msg, sizeof(msg));
	msg.key_msglen = sizeof(msg);
	msg.key_msgvers = KEY_VERSION;
	msg.key_msgtype = KEY_REGISTER;
	msg.type = type;

	if (write(fd, &msg, sizeof(msg)) < 0) {
		perror("write");
		return -1;
	}

	return 0;
}

int
init_kernelfd(void)
{
#ifdef NETKEY
	int s;

	if ((s = socket(PF_KEY, SOCK_RAW, 0)) < 0) {
		perror("socket");
		exit(1);
	}
	if (write_register(s, KEY_TYPE_AH) < 0) {
		exit(1);
	}
	if (write_register(s, KEY_TYPE_ESP) < 0) {
		exit(1);
	}

	return s;
#else /* NETKEY */
    struct sockaddr_in sin;
    int s;
    
    s = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (s == -1)
      exit_log("socket() in init_kernelfd()", 0, 0, 0);
    
    sin.sin_family = AF_INET;
    sin.sin_port = htons(our_port + 1);
    sin.sin_addr.s_addr = INADDR_ANY;

#ifdef DEBUG
    log(0, "init_kernelfd(): listening to port %d", our_port + 1, 0, 0);
#endif

    if (bind(s, (struct sockaddr *)&sin, sizeof(sin)) < 0)
      exit_log("bind() in init_kernelfd()", 0, 0, 0);
   
    return s;
#endif /* NETKEY */
}

/*
 * Initialize the sockets.
 */
int
init_socket(int *socks)
{
    struct protoent *proto;
    struct sockaddr_in sin, *sin2;
#ifdef INET6
    struct sockaddr_in6 sin6, *sin6p;
#endif /* INET6 */
    int sock, i, j, k;
    struct ifconf ifconf; 
    char buf[1024];

    /* Get a UDP socket */

    proto = getprotobyname("udp");
    
#ifdef INET6
    if (proto == (struct protoent *) NULL)
    {
	socks[0] = socket(PF_INET, SOCK_DGRAM, 0);
	socks[1] = socket(PF_INET6, SOCK_DGRAM, 0);
    }
    else
    {
	socks[0] = socket(PF_INET, SOCK_DGRAM, proto->p_proto);
	socks[1] = socket(PF_INET6, SOCK_DGRAM, proto->p_proto);
    }
    
    if (socks[0] == -1)
      exit_log("socket(PF_INET) failed in init_socket()", 0, 0, 0);
    if (socks[1] == -1)
      exit_log("socket(PF_INET6) failed in init_socket()", 0, 0, 0);
#else /* INET6 */
    if (proto == (struct protoent *) NULL)
      socks[0] = socket(PF_INET, SOCK_DGRAM, 0);
    else
      socks[0] = socket(PF_INET, SOCK_DGRAM, proto->p_proto);
    
    if (socks[0] == -1)
      exit_log("socket() failed in init_socket()", 0, 0, 0);

#endif /* INET6 */
#ifdef NETKEY
    k = 0;
#ifndef LIBRA
    if (setsockopt(socks[0], IPPROTO_IP, IP_AUTH_LEVEL, (void *)&k, 
		   sizeof(k)) < 0)
      exit_log("setsockopt(IPPROTO_IP)", 0, 0, 0);
#endif
#ifdef INET6
    if (setsockopt(socks[1], IPPROTO_IPV6, IPV6_AUTH_LEVEL, (void *)&k, 
		   sizeof(k)) < 0)
      exit_log("setsockopt(IPPROTO_IPV6)", 0, 0, 0);
#endif /* INET6 */
#endif /* NETKEY */
    sin.sin_family = AF_INET;
    sin.sin_port = htons(our_port);
    sin.sin_addr.s_addr = INADDR_ANY;

#ifdef INET6
    sin6.sin6_family = AF_INET6;
    sin6.sin6_port = htons(our_port);
    sin6.sin6_flowinfo = 0;
    sin6.sin6_addr = in6addr_any;

#endif /* INET6 */
#ifdef DEBUG
    log(0, "init_socket(): listening to port %d", our_port, 0, 0);
#endif

    k = 1;
    if (setsockopt(socks[0], SOL_SOCKET, SO_REUSEADDR, (void *)&k, 
		   sizeof(k)) < 0)
      exit_log("setsockopt() in init_socket()", 0, 0, 0);

    /* bind the socket */
    if (bind(socks[0], (struct sockaddr *)&sin, sizeof(sin)) < 0)
      exit_log("bind() failed in init_socket()", 0, 0, 0);

#ifdef INET6
    if (setsockopt(socks[1], SOL_SOCKET, SO_REUSEADDR, (void *)&k, 
		   sizeof(k)) < 0)
      exit_log("setsockopt() in init_socket()", 0, 0, 0);

    /* bind the socket */
    if (bind(socks[1], (struct sockaddr *)&sin6, sizeof(sin6)) < 0)
      exit_log("bind() failed in init_socket()", 0, 0, 0);
#endif /* INET6 */
    /* Get local interfaces */
    ifconf.ifc_len = 1024; 
    ifconf.ifc_buf = buf; 
    bzero(buf, 1024); 
 
    if (ioctl(socks[0], SIOCGIFCONF, &ifconf) == -1) 
      exit_log("ioctl() in init_socket()", 0, 0, 0); 

    /* This loop by Niels Provos */
#ifdef INET6
    for (i = 2, j = 0; j < ifconf.ifc_len; j += IFNAMSIZ + 
#else /* INET6 */
    for (i = 1, j = 0; j < ifconf.ifc_len; j += IFNAMSIZ + 
#endif /* INET6 */
#if defined(__NetBSD__) || defined(__OpenBSD__) || defined(__FreeBSD__)
	     buf[IFNAMSIZ + j] 
#else 
	     sizeof(struct sockaddr) 
#endif 
	     , i++)
    {
#ifdef INET6
	sin6p = (struct sockaddr_in6 *) &buf[IFNAMSIZ + j]; 
	if (sin6p->sin6_family == AF_INET6)
	{
	    if ((socks[i] = socket(PF_INET6, SOCK_DGRAM, proto->p_proto)) < 0)
		exit_log("socket() in init_socket()", 0, 0, 0); 

	    if (setsockopt(socks[i], SOL_SOCKET, SO_REUSEADDR, (void *)&k, 
		       sizeof(k)) < 0)
		exit_log("setsockopt() in init_socket()", 0, 0, 0);

	    sin6.sin6_port = htons(our_port);
	    sin6.sin6_flowinfo = 0;
	    sin6.sin6_addr = sin6p->sin6_addr; 
	    sin6.sin6_family = AF_INET6; 
	
	    if (bind(socks[i], (struct sockaddr *)&sin6, sizeof(sin6)) < 0)
		exit_log("bind() in init_socket()", 0, 0,0); 

#ifdef DEBUG
	{
	    char ntop_buf[INET6_ADDRSTRLEN];
	    inet_ntop(AF_INET6, &sin6.sin6_addr, ntop_buf, sizeof(ntop_buf));
	    log(0, "listening at %s", ntop_buf, 0, 0);
	}
#endif
	    continue;
	}
#endif /* INET6 */
	sin2 = (struct sockaddr_in *) &buf[IFNAMSIZ + j]; 
	
	if (sin2->sin_family != AF_INET)
	{
	    i--;
	    continue;
	}
	
	if ((socks[i] = socket(PF_INET, SOCK_DGRAM, proto->p_proto)) < 0)
	  exit_log("socket() in init_socket()", 0, 0, 0); 

	if (setsockopt(socks[i], SOL_SOCKET, SO_REUSEADDR, (void *)&k, 
		       sizeof(k)) < 0)
	  exit_log("setsockopt() in init_socket()", 0, 0, 0);

	bzero((void *)&sin, sizeof(sin));
	sin.sin_port = htons(our_port);
	sin.sin_addr = sin2->sin_addr; 
	sin.sin_family = AF_INET; 
	
	if (bind(socks[i], (struct sockaddr *)&sin, sizeof(sin)) < 0)
	  exit_log("bind() in init_socket()", 0, 0,0); 

#ifdef DEBUG
	log(0, "listening at %s", inet_ntoa(sin2->sin_addr), 0, 0);
#endif
    }
    
#ifdef DEBUG
    log(0, "listening at %d interfaces", i - 1, 0, 0);
#endif

    return i;
}

/*
 * This routine listens for incoming ISAKMP packets, kernel PF_KEY requests,
 * handles packet retransmitts and other events.
 */
void
call_server(void)
{
    struct sockaddr_in sin;
    int kernelfd;
    int sock[256];
    int numsocks;
    fd_set readfds;
    struct timeval tm;
    int i, maxfd, k;
    
    numsocks = init_socket(sock);
    kernelfd = init_kernelfd();

#ifdef DEBUG
    log(0, "socket numbers: ", 0, 0, 0);
    for (k = 0; k < numsocks; k++)
      fprintf(stderr, "%d ", sock[k]);
    fprintf(stderr, "\n");
    log(0, "kernel socket: %d", kernelfd, 0, 0);
#endif

    while (1)
    {
#ifdef INET6
	for (k = 2; k < numsocks; k++)
#else /* INET6 */
	for (k = 1; k < numsocks; k++)
#endif /* INET6 */
	  FD_SET(sock[k], &readfds);
	FD_SET(kernelfd, &readfds);

	maxfd = (sock[numsocks - 1] > kernelfd ? sock[numsocks - 1] : 
		 kernelfd) + 1;

	i = next_event();   /* Is there any pending event ? */
	if (i == -1)
	  i = select(maxfd, &readfds, NULL, NULL, NULL);
	else
	{
	    tm.tv_sec = i;
	    tm.tv_usec = 0;
	    i = select(maxfd, &readfds, NULL, NULL, &tm);
	}

	if (i == -1)
	  exit_log("select() failed in call_server()", 0, 0, 0);

#ifdef INET6
	for (k = 2; k < numsocks; k++)
#else /* INET6 */
	for (k = 1; k < numsocks; k++)
#endif /* INET6 */
	  if (FD_ISSET(sock[k], &readfds))
#ifdef DEBUG
	  {
	      log(0, "received packet", 0, 0, 0);
#endif
	      comm_handle(kernelfd, sock[k]);
#ifdef DEBUG
	  }
#endif

	if (FD_ISSET(kernelfd, &readfds))
#ifdef DEBUG
	{
	    log(0, "received kernel message", 0, 0, 0);
#endif
#ifdef INET6
	  kernel_handle(kernelfd, sock[0], sock[1]);
#else /* INET6 */
	  kernel_handle(kernelfd, sock[0]);
#endif /* INET6 */
#ifdef DEBUG
	}
#endif
	
	if (i == 0)   /* Event */
#ifdef DEBUG
	{
	    log(0, "time to handle event", 0, 0, 0);
#endif
#ifdef INET6
	  event_handle(kernelfd, sock[0], sock[1]);
#else /* INET6 */
	  event_handle(kernelfd, sock[0]);
#endif /* INET6 */
#ifdef DEBUG	    
	}
#endif
    }
}
