/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/socket.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include <netinet/in.h>
#ifdef INET6
#include <netinet6/in6.h>
#endif
#include <netkey/key.h>
#include <netinet6/ah.h>
#include <netinet6/esp.h>

static struct {
	char *name;
	int value;
	int vers;
	int defalgo;
} keytypetab[] = {
	{ "ah", KEY_TYPE_AH, 1, AH_ALGO_HMAC_MD5 },
	{ "esp", KEY_TYPE_ESP, 0, 0 },
	{ NULL, -1 },
};

static struct algotab {
	int keytype;
	char *algoname;
	int value;
	int (*check) __P((void));	/*to be used...*/
} algotab[] = {
	{ KEY_TYPE_AH, "new-hmac-md5", AH_ALGO_HMAC_MD5_NEW, NULL, },
	{ KEY_TYPE_AH, "new-hmac-sha1", AH_ALGO_HMAC_SHA1_NEW, NULL, },
	{ KEY_TYPE_AH, "new-none", AH_ALGO_NONE_NEW, NULL, },
	{ KEY_TYPE_AH, "hmac-md5", AH_ALGO_HMAC_MD5, NULL, },
	{ KEY_TYPE_AH, "hmac-sha1", AH_ALGO_HMAC_SHA1, NULL, },
	{ KEY_TYPE_AH, "keyed-md5", AH_ALGO_KEYED_MD5, NULL, },
	{ KEY_TYPE_AH, "md5", AH_ALGO_KEYED_MD5, NULL, },
	{ KEY_TYPE_AH, "none", AH_ALGO_NONE, NULL, },
	{ KEY_TYPE_ESP, "new-none", ESP_ALGO_NONE_NEW, NULL, },
	{ KEY_TYPE_ESP, "new-des", ESP_ALGO_DES_DERIVED, NULL, },
	{ KEY_TYPE_ESP, "new-des-expiv", ESP_ALGO_DES_EXPIV, NULL, },
	{ KEY_TYPE_ESP, "none", ESP_ALGO_NONE, NULL, },
	{ KEY_TYPE_ESP, "des-cbc", ESP_ALGO_DESCBC, NULL, },
	{ KEY_TYPE_ESP, "descbc", ESP_ALGO_DESCBC, NULL, },
	{ KEY_TYPE_ESP, "des", ESP_ALGO_DESCBC, NULL, },
	{ KEY_TYPE_ESP, "blowfish", ESP_ALGO_BLOWFISH_CBC, NULL, },
	{ KEY_TYPE_ESP, "blowfish-cbc", ESP_ALGO_BLOWFISH_CBC, NULL, },
	{ KEY_TYPE_ESP, "cast128-cbc", ESP_ALGO_CAST128CBC, },
	{ KEY_TYPE_ESP, "new-3des-expiv", ESP_ALGO_3DES_EXPIV, NULL, },
	{ KEY_TYPE_ESP, "rc5-cbc", ESP_ALGO_RC5_CBC, NULL, },
	{ -1, NULL },
};

static int verbose = 0;
static int keysock;
static enum { ADD, DELETE, FLUSH, DUMP, HEXDUMP, SCRIPTDUMP } mode = DUMP;
static int algorithm = AH_ALGO_KEYED_MD5;
static int keytype = KEY_TYPE_AH;
static int keyvers = 1;
static int af = AF_INET;
static int ivlength = 8;
static int hexkey = 0;
static int numeric = 0;

static void usage __P((void));
static void addkey __P((u_int, char *, char *, char *, char *, char *));
static void deletekey __P((u_int, char *, char *));
static void dumpkey __P((void));
static void flushkey __P((void));
static char *dumpsa __P((struct sockaddr *, int));
static void hex2bin __P((u_char *, size_t *, u_char *, size_t));
static void hexdump __P((FILE *, caddr_t, size_t));
static void keydump __P((FILE *, caddr_t, size_t));
static int setalgorithm __P((char *));
static char *getalgoname __P((int, int));
static int setkeytype __P((char *, int *));
static int setkeyvers __P((char *, int *));
static char *getkeytypename __P((int, int));

void
main(argc, argv)
	int argc;
	char **argv;
{
	char *s;

	argc--; argv++;

	/* first item: mode */
	if (argc) {
		if (strcmp(argv[0], "add") == 0)
			mode = ADD;
		else if (strcmp(argv[0], "delete") == 0)
			mode = DELETE;
		else if (strcmp(argv[0], "flush") == 0)
			mode = FLUSH;
		else if (strcmp(argv[0], "dump") == 0)
			mode = DUMP;
		else if (strcmp(argv[0], "hexdump") == 0)
			mode = HEXDUMP;
		else if (strcmp(argv[0], "scriptdump") == 0)
			mode = SCRIPTDUMP;
		else {
			usage();
			exit(1);
			/*NOTREACHED*/
		}

		argc--; argv++;
	}

	/* optional args */
	while (argc > 0 && **argv == '-') {
		for (s = argv[0] + 1; *s != '\0'; s++) {
			switch(*s) {
			case 'a':
			    {
				char *algo;

				if (s[1]) {
					algo = &s[1];
					while (*s)
						s++;
					s--;
				} else if (2 <= argc) {
					argc--; argv++;
					s = argv[0];
					algo = &s[0];
					while (*s)
						s++;
					s--;
				} else {
					usage();
					exit(1);
					/*NOTREACHED*/
				}

				algorithm = setalgorithm(algo);
				if (algorithm == -1) {
					fprintf(stderr,
						"invalid algorithm %s\n", algo);
					exit(1);
					/*NOTREACHED*/
				}
				break;
			    }

			case 'A':
			    {
				char *afstr;

				if (s[1]) {
					afstr = &s[1];
					while (*s)
						s++;
					s--;
				} else if (2 <= argc) {
					argc--; argv++;
					s = argv[0];
					afstr = &s[0];
					while (*s)
						s++;
					s--;
				} else {
					usage();
					exit(1);
					/*NOTREACHED*/
				}

				if (strcmp(afstr, "inet") == 0)
					af = AF_INET;
#ifdef INET6
				else if (strcmp(afstr, "inet6") == 0)
					af = AF_INET6;
#endif
				else {
					fprintf(stderr,
						"invalid AF %s\n", afstr);
					exit(1);
					/*NOTREACHED*/
				}
				break;
			    }

			case 't':
			    {
				char *type;

				if (s[1]) {
					type = &s[1];
					while (*s)
						s++;
					s--;
				} else if (2 <= argc) {
					argc--; argv++;
					s = argv[0];
					type = &s[0];
					while (*s)
						s++;
					s--;
				} else {
					usage();
					exit(1);
					/*NOTREACHED*/
				}

				keytype = setkeytype(type, &algorithm);
				keyvers = setkeyvers(type, &algorithm);
				if (keytype == -1 || keyvers == -1) {
					fprintf(stderr,
						"invalid keytype %s\n", type);
					exit(1);
					/*NOTREACHED*/
				}
				break;
			    }

			case 'i':
			    {
				char *istr;

				if (s[1]) {
					ivlength = atoi(&s[1]);
					while (*s)
						s++;
					s--;
				} else if (2 <= argc) {
					argc--; argv++;
					s = argv[0];
					ivlength = atoi(s);
					while (*s)
						s++;
					s--;
				} else {
					usage();
					exit(1);
					/*NOTREACHED*/
				}
				break;
			    }

			case 'h':
				hexkey++;
				break;
			case 'n':
				numeric++;
				break;
			case 'v':
				verbose++;
				break;

			default:
				fprintf(stderr, "invalid option -%c\n", *s);
				exit(1);
				/*NOTREACHED*/
			}
		}
		argc--; argv++;
	}

    {
	time_t t;
	t = time(NULL);
	t ^= (getpid() & 0xffffffff);
	srand(t & 0xffffffff);
    }

	keysock = socket(PF_KEY, SOCK_RAW, af);
	if (keysock < 0)
		perror("socket(PF_KEY)");

	/* rest of args */
	switch (mode) {
	case ADD:
	    {
		char *key;
		char *ekey;

		if (argc != 4 && argc != 5 && argc != 6) {
			usage();
			exit(1);
			/*NOTREACHED*/
		}
		key = ekey = NULL;
		if (5 <= argc)
			key = argv[4];
		if (6 <= argc)
			ekey = argv[5];
		if (verbose) {
			fprintf(stderr,
				"adding key spi=%s algo=%d(%s) src=%s dst=%s from=%s",
				argv[0], algorithm,
				getalgoname(algorithm, keytype),
				argv[1], argv[2], argv[3]);
			if (key)
				fprintf(stderr, ", key=<%s>", key);
			if (ekey)
				fprintf(stderr, ", ekey=<%s>", ekey);
			fprintf(stderr, "\n");
		}

		if (algorithm == ESP_ALGO_CAST128CBC) {
			size_t keylen;

			ivlength = 8; /* 8 Byte only */

			if (argc < 5) {
				fprintf(stderr, "CAST128-CBC require "
						"128 bit key\n");
				exit(EXIT_FAILURE);
			}

			keylen = strlen(argv[4]);
			if ((keylen < 40/8) || (keylen > 128/8)) {
				fprintf(stderr, "Error: cast128 key "
					"need 40 to 128 bits\n");
				exit(EXIT_FAILURE);
			}
		} else if (algorithm == ESP_ALGO_3DES_EXPIV) {
			size_t keylen;

			if (ivlength != 8) {
				fprintf(stderr, "Sorry, new-3des-expiv is"
						"only 8 byte IV\n");
				ivlength = 8;
			}
			keylen = strlen(argv[4]);
			if (keylen != 24) {
				fprintf(stderr, "new-3des-expiv require "
						"192 bits key\n");
				exit(EXIT_FAILURE);
			}
			if (memcmp(argv[4], argv[4]+8, 8) == 0) {
				fprintf(stderr, "new-3des-expiv require "
						"that first key is not "
						"equal to second key\n");
				exit(EXIT_FAILURE);
			}
		} else if (algorithm == ESP_ALGO_RC5_CBC) {
			size_t keylen;

			keylen = strlen(argv[4]);
			if ((keylen < 40/8) || (keylen > 2040/8)) {
				fprintf(stderr, "Error: rc5-cbc "
					"key need 40 to 2040 bits\n");
				exit(EXIT_FAILURE);
			}
		}

		addkey(atoi(argv[0]), argv[1], argv[2], argv[3],
			key ? key : "", ekey ? ekey : "");
	    }
		break;

	case DELETE:
		if (argc != 3) {
			usage();
			exit(1);
			/*NOTREACHED*/
		}
		if (verbose) {
			fprintf(stderr,
				"deleteing key spi=%s type=%d(%s) src=%s dst=%s",
				argv[0], keytype,
				 getkeytypename(keytype, keyvers),
				argv[1], argv[2]);
			fprintf(stderr, "\n");
		}
		deletekey(atoi(argv[0]), argv[1], argv[2]);
		break;

	case FLUSH:
		if (argc != 0) {
			usage();
			exit(1);
			/*NOTREACHED*/
		}
		if (verbose)
			fprintf(stderr, "flushing key table\n");
		flushkey();
		break;

	case DUMP:
	case HEXDUMP:
	case SCRIPTDUMP:
		if (argc != 0) {
			usage();
			exit(1);
			/*NOTREACHED*/
		}
		if (verbose)
			fprintf(stderr, "dumping key table\n");
		dumpkey();
		break;
	}

	exit(0);
}

static void
usage()
{
	fprintf(stderr,
"usage:\n"
"\tsetkey add [options] spi src dst from [key [ekey]]\n"
"\tsetkey delete [options] spi src dst\n"
"\tsetkey flush [options]\n"
"\tsetkey dump [options]\n"
"\tsetkey hexdump [options]\n"
"\tsetkey scriptdump [options]\n"
);
}

static void
addkey(spi, srcname, dstname, fromname, keystr, ekeystr)
	u_int spi;
	char *srcname;
	char *dstname;
	char *fromname;
	char *keystr;	/*null-terminated string, which is evil*/
	char *ekeystr;	/*null-terminated string, which is evil*/
{
	u_char buf[BUFSIZ];
	u_char keybuf[BUFSIZ];
	u_char ekeybuf[BUFSIZ];
	struct key_msghdr *msg;
	struct sockaddr_in *src;
	struct sockaddr_in *dst;
	struct sockaddr_in *from;
#ifdef INET6
	struct sockaddr_in6 *src6;
	struct sockaddr_in6 *dst6;
	struct sockaddr_in6 *from6;
#endif
	u_char *key;
	u_char *ekey;
	u_char *iv;
	size_t keylen;
	size_t ekeylen;
	size_t ivlen;
	int len;

	if (hexkey) {
		hex2bin(keybuf, &keylen, keystr, strlen(keystr));
		keystr = &keybuf[0];
		hex2bin(ekeybuf, &ekeylen, ekeystr, strlen(ekeystr));
		ekeystr = &ekeybuf[0];
	} else {
		keylen = strlen(keystr);	/*XXX zero terminate?*/
		ekeylen = strlen(ekeystr);	/*XXX zero terminate?*/
	}
	ivlen = ivlength;

	/*
	 * perform special checks for HMAC-*-96 algorithms.
	 * XXX should be generalized
	 */
	if (keylen != 128/8
	 && keytype == KEY_TYPE_AH
	 && (algorithm == AH_ALGO_HMAC_MD5_NEW
	  || algorithm == AH_ALGO_HMAC_MD5)) {
		fprintf(stderr, "bad key length for hmac-md5 "
			"(must be 128 bits)\n");
		return;
	}
	if (keylen != 160/8
	 && keytype == KEY_TYPE_AH
	 && (algorithm == AH_ALGO_HMAC_SHA1_NEW
	  || algorithm == AH_ALGO_HMAC_SHA1)) {
		fprintf(stderr, "bad key length for hmac-sha1 "
			"(must be 160 bits)\n");
		return;
	}

	len = 0;
	msg = (struct key_msghdr *)&buf[0];
	len += sizeof(struct key_msghdr);
#ifdef INET6
	if (af == AF_INET6) {
		src6 = (struct sockaddr_in6 *)(msg + 1);
		len += sizeof(struct sockaddr_in6);
		dst6 = (struct sockaddr_in6 *)(src6 + 1);
		len += sizeof(struct sockaddr_in6);
		from6 = (struct sockaddr_in6 *)(dst6 + 1);
		len += sizeof(struct sockaddr_in6);
		key = (u_char *)(from6 + 1);
	} else
#endif
	if (af == AF_INET) {
		src = (struct sockaddr_in *)(msg + 1);
		len += sizeof(struct sockaddr_in);
		dst = (struct sockaddr_in *)(src + 1);
		len += sizeof(struct sockaddr_in);
		from = (struct sockaddr_in *)(dst + 1);
		len += sizeof(struct sockaddr_in);
		key = (u_char *)(from + 1);
	} else
		abort();
	len += keylen;
	iv = (u_char *)(key + keylen);
	len += ivlen;
	ekey = (u_char *)(iv + ivlen);
	len += ekeylen;

	bzero(buf, len);

	msg->key_msglen = len;
	msg->key_msgvers = KEY_VERSION;
	msg->key_msgtype = KEY_ADD;
	msg->type = keytype;
	msg->vers = keyvers;
	msg->state = 0;
	msg->label = 0;
	msg->spi = spi;
	msg->keylen = keylen;
	msg->ekeylen = ekeylen;
	msg->ivlen = ivlen;
	msg->algorithm = algorithm;
	msg->lifetype = 0;	/*XXX*/
	msg->lifetime1 = 3600;	/*XXX*/
	msg->lifetime2 = 3600;	/*XXX*/
	if (keytype == KEY_TYPE_AH)
		msg->antireplay = 1;
	else if (keytype == KEY_TYPE_ESP && ekeylen)
		msg->antireplay = 1;
	else
		msg->antireplay = 0;
#ifdef INET6
	if (af == AF_INET6) {
		bzero(src6, sizeof(struct sockaddr_in6));
		src6->sin6_len = sizeof(struct sockaddr_in6);
		src6->sin6_family = af;
		(void)inet_pton(af, srcname, &src6->sin6_addr);
		bzero(dst6, sizeof(struct sockaddr_in6));
		dst6->sin6_len = sizeof(struct sockaddr_in6);
		dst6->sin6_family = af;
		(void)inet_pton(af, dstname, &dst6->sin6_addr);
		bzero(from6, sizeof(struct sockaddr_in6));
		from6->sin6_len = sizeof(struct sockaddr_in6);
		from6->sin6_family = AF_INET6;
		(void)inet_pton(af, fromname, &from6->sin6_addr);
	} else
#endif
	if (af == AF_INET) {
		bzero(src, sizeof(struct sockaddr_in));
		src->sin_len = sizeof(struct sockaddr_in);
		src->sin_family = af;
		(void)inet_pton(af, srcname, &src->sin_addr);
		bzero(dst, sizeof(struct sockaddr_in));
		dst->sin_len = sizeof(struct sockaddr_in);
		dst->sin_family = af;
		(void)inet_pton(af, dstname, &dst->sin_addr);
		bzero(from, sizeof(struct sockaddr_in));
		from->sin_len = sizeof(struct sockaddr_in);
		from->sin_family = af;
		(void)inet_pton(af, fromname, &from->sin_addr);
	} else
		abort();
	bzero(key, keylen);
	bzero(iv, ivlen);
	bcopy(keystr, key, keylen);
    {
	int i;

	for (i = 0; i < ivlen; i++)
		iv[i] = rand() & 0xff;
    }
	bcopy(ekeystr, ekey, ekeylen);

	if (write(keysock, &buf[0], len) < 0) {
		perror("write(keysock, KEY_ADD)");
		return;
	}
	if (read(keysock, &buf[0], BUFSIZ) < 0) {
		perror("read(keysock, KEY_ADD)");
		return;
	}
	if (msg->key_errno != 0)
		fprintf(stderr, "key_errno=%d\n", msg->key_errno);
}

static void
deletekey(spi, srcname, dstname)
	u_int spi;
	char *srcname;
	char *dstname;
{
	u_char buf[BUFSIZ];
	struct key_msghdr *msg;
	struct sockaddr_in *src;
	struct sockaddr_in *dst;
#ifdef INET6
	struct sockaddr_in6 *src6;
	struct sockaddr_in6 *dst6;
#endif
	int len;

	len = 0;
	msg = (struct key_msghdr *)&buf[0];
	len += sizeof(struct key_msghdr);
#ifdef INET6
	if (af == AF_INET6) {
		src6 = (struct sockaddr_in6 *)(msg + 1);
		len += sizeof(struct sockaddr_in6);
		dst6 = (struct sockaddr_in6 *)(src6 + 1);
		len += sizeof(struct sockaddr_in6);
	} else
#endif
	if (af == AF_INET) {
		src = (struct sockaddr_in *)(msg + 1);
		len += sizeof(struct sockaddr_in);
		dst = (struct sockaddr_in *)(src + 1);
		len += sizeof(struct sockaddr_in);
	} else
		abort();

	bzero(buf, len);

	msg->key_msglen = len;
	msg->key_msgvers = KEY_VERSION;
	msg->key_msgtype = KEY_DELETE;
	msg->type = keytype;
	msg->vers = keyvers;
	msg->spi = spi;
#ifdef INET6
	if (af == AF_INET6) {
		bzero(src6, sizeof(struct sockaddr_in6));
		src6->sin6_len = sizeof(struct sockaddr_in6);
		src6->sin6_family = af;
		(void)inet_pton(af, srcname, &src6->sin6_addr);
		bzero(dst6, sizeof(struct sockaddr_in6));
		dst6->sin6_len = sizeof(struct sockaddr_in6);
		dst6->sin6_family = af;
		(void)inet_pton(af, dstname, &dst6->sin6_addr);
	} else
#endif
	if (af == AF_INET) {
		bzero(src, sizeof(struct sockaddr_in));
		src->sin_len = sizeof(struct sockaddr_in);
		src->sin_family = af;
		(void)inet_pton(af, srcname, &src->sin_addr);
		bzero(dst, sizeof(struct sockaddr_in));
		dst->sin_len = sizeof(struct sockaddr_in);
		dst->sin_family = af;
		(void)inet_pton(af, dstname, &dst->sin_addr);
	} else
		abort();

	if (write(keysock, &buf[0], len) < 0) {
		perror("write(keysock, KEY_DELETE)");
		return;
	}
	if (read(keysock, &buf[0], BUFSIZ) < 0) {
		perror("read(keysock, KEY_DELETE)");
		return;
	}
	if (msg->key_errno != 0)
		fprintf(stderr, "key_errno=%d\n", msg->key_errno);
}

static void
dumpkey()
{
	u_char buf[BUFSIZ];
	struct key_msghdr *msg;
	int len;
	struct sockaddr *src;
	struct sockaddr *dst;
	struct sockaddr *from;
	u_char *key;
	u_char *ekey;
	u_char *iv;
	u_char tmp[BUFSIZ];

	len = 0;
	msg = (struct key_msghdr *)&buf[0];
	len += sizeof(struct key_msghdr);
	msg->key_msglen = len;
	msg->key_msgvers = KEY_VERSION;
	msg->key_msgtype = KEY_DUMP;

	if (write(keysock, &buf[0], len) < 0) {
		perror("write(keysock, KEY_DUMP)");
		return;
	}

	if (mode == DUMP) {
		printf("seq spi     type    algo      src             dst             from\n");
	}

	while (1) {
		len = read(keysock, &buf[0], BUFSIZ);
		if (len < 0) {
			perror("read(keysock)");
			return;
		}
		if (msg->key_errno != 0)
			fprintf(stderr, "ERR: key_errno=%d\n", msg->key_errno);

		if (len != msg->key_msglen) {
			fprintf(stderr, "ERR: size mismatch? %d %d\n",
				len, msg->key_msglen);
			return;
		}

		src = (struct sockaddr *)(msg + 1);
		dst = (src && src->sa_len)
			? (struct sockaddr *)(((u_char *)src) + src->sa_len)
			: NULL;
		from = (dst && dst->sa_len)
			? (struct sockaddr *)(((u_char *)dst) + dst->sa_len)
			: NULL;
		key = (from && from->sa_len)
			? (((u_char *)from) + from->sa_len)
			: NULL;
		iv = key ? (key + msg->keylen) : NULL;
		ekey = iv ? (iv + msg->ivlen) : NULL;

		/*
		 * hex dump
		 */
		if (mode == HEXDUMP) {
			if (msg->key_seq == 0) {
				printf("<end>\n");
				return;
			}

			printf("<seq=%d>\n", msg->key_seq);
			hexdump(stdout, buf, len);
			continue;
		}

		/*
		 * script dump
		 */
		if (mode == SCRIPTDUMP) {
			size_t i;

			if (msg->key_seq == 0)
				return;
			printf("setkey add -h ");
			printf("-t %s ", getkeytypename(msg->type, msg->vers));
			printf("-a %s ",
				getalgoname(msg->algorithm, msg->type));
			printf("-A %s ",
				(src->sa_family == AF_INET) ? "inet" : "inet6");
			printf("%u %s ",
				msg->spi, dumpsa(src, BUFSIZ));
			printf("%s ", dumpsa(dst, BUFSIZ));
			printf("%s ", dumpsa(from, BUFSIZ));
			for (i = 0; i < msg->keylen; i++)
				printf("%02x", key[i]);
			printf(" ");
			for (i = 0; i < msg->ekeylen; i++)
				printf("%02x", ekey[i]);
			printf("\n");
			continue;
		}

		/*
		 * verbal dump
		 */
		if (msg->key_seq == 0) {
			if (verbose)
				printf("end\n");
			return;
		}
		printf("%-4d", msg->key_seq);

		if (msg->key_msgvers != KEY_VERSION) {
			printf("(unknown version %d)\n",
				msg->key_msgvers);
			continue;
		}

		printf("%-8u", msg->spi);

		if (!numeric && getkeytypename(msg->type, msg->vers))
			printf("%-8s", getkeytypename(msg->type, msg->vers));
		else
			printf("%-8d", msg->type);

		if (!numeric && getalgoname(msg->algorithm, msg->type)) {
			if (8 < strlen(tmp))
				tmp[8] = '\0';
			printf("%-10s", getalgoname(msg->algorithm, msg->type));
		} else
			printf("%-10d", msg->algorithm);

#if 0
		printf("%-8d", msg->state);
		printf("%-8d", msg->label);

		printf("%-8d%-8d%-8d",
			msg->lifetype, msg->lifetime1, msg->lifetime2);
#endif

		printf("%-16s", src ? dumpsa(src, 16) : "(none)");
		printf("%-16s", dst ? dumpsa(dst, 16) : "(none)");
		printf("%-16s", from ? dumpsa(from, 16) : "(none)");

		printf("\n");

		if (msg->keylen) {
			printf("  key: len=%d, ", msg->keylen);
			keydump(stdout, key, msg->keylen);
			printf("\n");
		}

		if (msg->ivlen) {
			printf("  iv: len=%d, ", msg->ivlen);
			keydump(stdout, iv, msg->ivlen);
			printf("\n");
		}

		if (msg->ekeylen) {
			printf("  ekey: len=%d, ", msg->ekeylen);
			keydump(stdout, ekey, msg->ekeylen);
			printf("\n");
		}
	}
}

static void
flushkey()
{
	u_char buf[BUFSIZ];
	struct key_msghdr *msg;
	int len;
	struct sockaddr *sa;

	len = 0;
	msg = (struct key_msghdr *)&buf[0];
	len += sizeof(struct key_msghdr);
	msg->key_msglen = len;
	msg->key_msgvers = KEY_VERSION;
	msg->key_msgtype = KEY_FLUSH;

	if (write(keysock, &buf[0], len) < 0) {
		perror("write(keysock, KEY_FLUSH)");
		return;
	}
	if (read(keysock, &buf[0], BUFSIZ) < 0) {
		perror("read(keysock, KEY_FLUSH)");
		return;
	}
	if (msg->key_errno != 0)
		fprintf(stderr, "key_errno=%d\n", msg->key_errno);
}

static char *
dumpsa(sa, len)
	struct sockaddr *sa;
	int len;
{
#ifdef INET6
	static char ntop_buf[INET6_ADDRSTRLEN];
	struct sockaddr_in *sin;
	struct sockaddr_in6 *sin6;

	switch (sa->sa_family) {
	case AF_INET6:
		sin6 = (struct sockaddr_in6 *)sa;
		inet_ntop(sin6->sin6_family,
			&sin6->sin6_addr, ntop_buf, sizeof(ntop_buf));
		if (len < sizeof(ntop_buf))
			ntop_buf[len] = '\0';
		return ntop_buf;

	case AF_INET:
		sin = (struct sockaddr_in *)sa;
		inet_ntop(sin->sin_family,
			&sin->sin_addr, ntop_buf, sizeof(ntop_buf));
		if (len < sizeof(ntop_buf))
			ntop_buf[len] = '\0';
		return ntop_buf;

	default:
		return "(AF?)";
	}
#else
	struct sockaddr_in *sin;
	char *p;

	switch (sa->sa_family) {
	case AF_INET:
		sin = (struct sockaddr_in *)sa;
		p = inet_ntoa(sin->sin_addr);
		if (len < strlen(p))
			p[len] = '\0';
		return p;

	default:
		return "(AF?)";
	}
#endif
}

static void
hex2bin(dst, dstlen, src, srclen)
	u_char *dst;
	size_t *dstlen;
	u_char *src;
	size_t srclen;
{
	size_t i;

#define	ishex(x)	\
    (isdigit(x) || ('a' <= (x) && (x) <= 'f') || ('A' <= (x) && (x) <= 'F'))
#define	hexch(x)	\
    (isdigit(x) ? ((x) - '0') : ((x) & 0x4f) - 'A' + 10)

	/*sanity*/
	if (srclen % 2) {
		fprintf(stderr,
			"strange digit count on hex string\n");
		exit(1);
	}
	for (i = 0; i < srclen; i++) {
		if (!ishex(src[i])) {
			fprintf(stderr,
				"non-hex character appear on hex string\n");
			exit(1);
		}
	}

	*dstlen = 0;
	for (i = 0; i < srclen; i += 2) {
		dst[*dstlen] = hexch(src[i]) << 4;
		dst[*dstlen] |= hexch(src[i + 1]);
		*dstlen = *dstlen + 1;
	}
#undef ishex
#undef hexch
}

/*------------------------------------------------------------*/

void
hexdump(f, buf, len)
	FILE *f;
	caddr_t buf;
	size_t len;
{
	int i;
	u_char *p;

	p = buf;
	for (i = 0; i < len; i++) {
		fprintf(f, "%02x ", p[i]);
		if (i % 16 == 15)
			fprintf(f, "\n");
	}
	if (len % 16)
		fprintf(f, "\n");
}

static void
keydump(f, buf, len)
	FILE *f;
	caddr_t buf;
	size_t len;
{
	int i;
	u_char *p;

	p = buf;
	fprintf(f, "<");
	for (i = 0; i < len; i++)
		fprintf(f, "%s%02x", (i != 0) ? " " : "", p[i]);
	fprintf(f, ">");
}

static int
setalgorithm(algo)
	char *algo;
{
	size_t i;

	i = 0;
	while (algotab[i].algoname) {
		if (algotab[i].keytype == keytype
		 && strcmp(algotab[i].algoname, algo) == 0)
			return algotab[i].value;
		i++;
	}

	return -1;
}

static char *
getalgoname(algoidx, keytype)
	int algoidx;
	int keytype;
{
	size_t i;

	i = 0;
	while (algotab[i].algoname) {
		if (algotab[i].keytype == keytype
		 && algotab[i].value == algoidx)
			return algotab[i].algoname;
		i++;
	}

	return NULL;
}

static int
setkeytype(keytype, palgo)
	char *keytype;
	int *palgo;
{
	size_t i;

	i = 0;
	while (keytypetab[i].name) {
		if (strcmp(keytypetab[i].name, keytype) == 0) {
			*palgo = keytypetab[i].defalgo;
			return keytypetab[i].value;
		}
		i++;
	}

	return -1;
}

static int
setkeyvers(keytype, palgo)
	char *keytype;
	int *palgo;
{
	size_t i;

	i = 0;
	while (keytypetab[i].name) {
		if (strcmp(keytypetab[i].name, keytype) == 0) {
			*palgo = keytypetab[i].defalgo;
			return keytypetab[i].vers;
		}
		i++;
	}

	return -1;
}

static char *
getkeytypename(keytypeidx, versidx)
	int keytypeidx;
	int versidx;
{
	size_t i;

	i = 0;
	while (keytypetab[i].name) {
		if (keytypetab[i].value == keytypeidx
		 && keytypetab[i].vers == versidx)
			return keytypetab[i].name;
		i++;
	}

	return NULL;
}
